package HTML::Mason::Forms;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);
use Data::Dumper;
use URI;

require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
	preprocess
);
$VERSION = '0.02';

sub DEBUG () {0}

#########################################################
use  vars qw($session $session_load $session_save $err_missing $err_bad
             %dispatch);
%dispatch=(
'MASON'=>\&mason,
'FORM'=>\&form,
'/FORM'=>\&form_END,
'INPUT'=>\&input,
'SELECT'=>\&__select,
'/SELECT'=>\&select_END,
'TEXTAREA'=>\&textarea,
'OPTION'=>\&option,
'NEXT'=>\&__next,
'BYPASS'=>\&bypass,
);

#########################################################
sub preprocess
{
    my($sref)=@_;

    $session=0;
    $session_load='';
    $session_save='';
    $err_missing="Missing field(s): ";
    $err_bad="Bad field(s): ";

    local $Data::Dumper::Purity=1;
    _clear_box();
    _clear_select();
    my $change=1 if $$sref=~s(<  
                        (?: (?:%(MASON|INPUT|SELECT|OPTION|FORM|NEXT|BYPASS)) 
                        |
                        (?:/(%SELECT|%FORM))
                    )
                    (.*?)
                 >)
                (_preprocess(uc($1||$2), $3))gisex;

    $change=1 if $$sref=~s(<
                    %(TEXTAREA)(.*?)>(.*?)</%TEXTAREA>
                )
                (_preprocess(uc($1), $2, $3))gisex;

    if($change)
    {
        select_END();         # in case the users wasn't paying attention...
        finish_boxes();       # close up all check and radio boxes
        $$sref=_perl_args()._submit()._session_var()._validation().$$sref._tailing();
        return 1;
    }
    return 0;
}

#########################################################
sub _preprocess
{
    my($tag, $atts, $extra)=@_;
    $tag=~s(^%)(/);

    my %atts=parse_atts($atts);
    my $ret="<$tag$atts>";

    $ret=$dispatch{$tag}->(\%atts, $ret, $extra) if exists $dispatch{$tag};

    return $ret;
}



#########################################################
{
    my @perl_args;
    sub _perl_args
    {
        if(@_)
        {
            DEBUG and warn "args:", join ', ', grep defined, @_;
            push @perl_args, grep defined, @_;
        } elsif(@perl_args)
        {
            my $t=qq(<%args>\n).join "\n", @perl_args, qq(</%args>);
            @perl_args=();
            return $t;
        } else
        {
            return '';
        }
    }
}

#########################################################
{
    my @submit;
    sub _submit
    {
        if(@_)
        {
            push @submit, grep defined, @_;
        } elsif(@submit)
        {
            my $t=qq[<%init>\nmy \$FIRST=not (].
                  join(" or ", map {qq(exists \$ARGS{$_})} @submit).
                  qq[);\n</%init>];
            @submit=();
            return $t;
        } else
        {
            return qq(<%init>my \$FIRST=1;\n</%init>);
        }
    }
}
#########################################################
{
    my @tailing;
    sub _tailing
    {
        if(@_)
        {
            push @tailing, grep defined, @_;
        } elsif(@tailing)
        {
            my $t=qq[<%init>].join("\n", @tailing).qq[\n</%init>];
            @tailing=();
            return $t;
        } else
        {
            return "";
        }
    }
}

#########################################################
{
    my @session_vars;
    sub _session_var
    {
        return '' unless $session;
        my $t='';
        my $cleanup='';
        if(@_)
        {
            push @session_vars, grep defined, @_;
        } elsif(@session_vars)
        {
            my($q, $s, %done);
            my $init='';
            foreach my $var (@session_vars)
            {
                next if $done{$var}++;
                $s=substr($var, 0, 1);
                $q=substr($var, 1);

                if($s eq '$'){
                    $init.=qq(    $var = \$SESSION{'$q'}||'' if exists \$SESSION{'$q'};)
                } else {
                    $init.=qq(    $var = $s\{\$SESSION{'$q'}} if exists \$SESSION{'$q'};)
                }
                $init.="\n";
                
                $cleanup.=qq(        \$SESSION{'$q'}=) . ($s eq '$' ? '' : '\\') . qq($var;\n);
            }

            @session_vars=();
            $t=<<MASON if $init;
<%init>
$session_load
if(\$FIRST and defined \%SESSION) {
$init
}
</%init>
MASON
        }

        $t.=<<MASON;
<%init>
my \$SAVE_SESSION=sub {
    if(defined \%SESSION) {
$cleanup
        tied(%SESSION)->save() if tied(%SESSION) and tied(%SESSION)->can('save');
    }
};
</%init>
<%cleanup>
&\$SAVE_SESSION if defined \%SESSION;
$session_save
</%cleanup>
MASON
        return $t;
    }
}

#########################################################
{
    my @validation;
    sub _validation
    {
        if(@_)
        {
            push @validation, grep defined, @_;
        } elsif(@validation)
        {
            my $v=join "\n    ", @validation;
            my $t=<<MASON;
<%init>
my \$VALIDATION_ERROR='';
unless(\$FIRST) {
    my(\@err, \@missing, \@bad);
    local \$SIG{__DIE__}=sub {push \@err, join '', \@_};
    $v
    push \@err, "$err_bad ".join ', ', \@bad if \@bad;
    push \@err, "$err_missing ".join ', ', \@missing if \@missing;
    \$VALIDATION_ERROR=join '<BR>', \@err if \@err;
}
</%init>
MASON
            @validation=();
            return $t;
        } else
        {
            return qq(<%INIT>my \$VALIDATION_ERROR=0;\n</%INIT>);
        }
    }
}

#########################################################
sub mason
{
    my($atts)=@_;
    my $src=$atts->{src};
    unless($src) {
        # warn "<%MASON> tag doesn't have a SRC attribute" unless $src;
        die "<%MASON> tag doesn't have a SRC attribute" unless $src;
    }
    delete $atts->{src};
#    local $Data::Dumper::Useqq=1;
    my $a=Data::Dumper->Dump([$atts], [qw(*my_args)]);
    $a=~s('\$\{(\w+?)\}')(\$$1)g;
    return join '', "<%perl>{my ${a}mc_comp('$src', \%my_args);}</%perl>";
}

#########################################################
{
    my(%checkes, %radios);
    sub _clear_box
    {
        %checkes=();
        %radios=();
    }

    #########################################################
    sub finish_boxes
    {
        local $_;
        my($a, @one);
        # Create <%perl_args> for the check boxes

        foreach my $name (keys %checkes)
        {
            # multiple checkboxes becomes a hash
            @{$checkes{$name}}=grep {$_} @{$checkes{$name}}; 
            $a=Data::Dumper->Dump([{map {$_ => 1} @{$checkes{$name}} }],
                                  ["*$name"]);
            $a=~s/\s+/ /g;
            $a=~s/ = / => /;
            $a=~s/;\s*$//;
            _perl_args($a);
        }

        # Create <%perl_args> for the radio buttons
        foreach my $name (keys %radios)
        {
            _perl_args(qq(\$$name=>").($radios{$name}||'').q("));
        }
    }

    #########################################################
    sub input
    {
        my($atts, $ret)=@_;
        $atts->{type}||='TEXT';
        my $type=uc $atts->{type};

        unless($atts->{name}) {
            warn qq(<%INPUT type=$type> without a name\n) unless $atts->{name};
            return $ret;
        }
        my $name=_clean_name($atts->{name});
        my $value=quotemeta($atts->{value}||'');

        my $boxes;
        my $s='$';
        if($type eq 'TEXT' or $type eq 'PASSWORD' or
           $type eq 'HIDDEN')
        {
            delete @{$atts}{qw(value)};
            _perl_args(qq(\$$name=>"$value"));
            $ret=qq(<INPUT value="<%\$$name%>");
        } elsif($type eq 'RADIO')
        {
            if(exists $atts->{checked})
            {
                $radios{$name}||=$value;
            } elsif(not exists $radios{$name})
            {
                $radios{$name}='';
            }

            delete @{$atts}{qw(checked)};
            $ret=qq(<INPUT<%(\$$name eq "$value" ? ' CHECKED' : '')%>);
        } elsif($type eq 'CHECKBOX')
        {
            if(exists $atts->{checked})
            {
                push @{$checkes{$name}}, $value
            } elsif(not exists $checkes{$name})
            {
                push @{$checkes{$name}}, '';
            }
            $s='%';

            delete @{$atts}{qw(checked)};
            $ret=qq(<INPUT<%(exists \$$name\{"$value"} ? ' CHECKED' : '')%>);
        } elsif($type eq 'SUBMIT')
        {
            _submit($name);
            return $ret;
        } else
        {
            return $ret;
        }
        _std_var($s, $name, $atts);

        my $more=_nice_atts($atts);
        return $ret."$more>";
    }
}

#########################################################
{
    my($select, %chosen);
    sub _clear_select
    {
        $select='';
        %chosen=();
    }

    #########################################################
    sub _select
    {
        my($atts, $ret)=@_;
        if($select)
        {
            warn qq(Unclosed <%SELECT name="$select">\n);
            $ret=select_END($ret);
        }
        unless($atts->{name}) {
            warn qq(<%SELECT> without a name\n) unless $atts->{name};
            return $ret;
        }
        $select=_clean_name($atts->{name});
        %chosen=();
        return $ret;
    }

    #########################################################
    sub select_END
    {
        my($atts, $ret)=@_;
        return if not keys %chosen;

        my $a=Data::Dumper->Dump([\%chosen], ["*$select"]);
        $a=~s/\s+/ /g;
        $a=~s/ = / => /;
        _perl_args($a);

        _clear_select();
        return $ret;
    }

    #########################################################
    sub option
    {
        my($atts, $ret)=@_;
        my $value=quotemeta $atts->{value};
        my $t=yesno_att($atts, 'selected');

        $chosen{$value}=1 if $t;
        delete $atts->{selected};
        
        my $more=_nice_atts($atts);
        $ret=qq(<OPTION$more<%(exists(\$$select\{"$value"}) ? ' SELECTED' : '')%>>);
        return $ret;
    }

}

#########################################################
sub _std_var
{
    my($s, $name, $atts)=@_;;

    # create the session var
    _session_var("$s$name") unless yesno_att($atts, 'notsession');

    # do required fields
    my $desc=$atts->{desc}||$name;
    $desc=~s/'/\'/g;
    _validation(qq(push \@missing, '$desc' unless $s$name;))
                    if yesno_att($atts, 'required');

    # error message for field validation
    my $err=$atts->{error}||'';
    $err=~s/'/\'/g;
    $err= $err ? qq[push \@error, '$err'] : qq[push \@bad, '$desc'];

    # validation code
    _validation(_nice_code(qq($err unless $atts->{valid};)))
                    if $atts->{'valid'};
    # validation regex
    _validation(qq[$err unless $s$name=~m"$atts->{regex}";])
                    if $atts->{'regex'};

    # get rid of attributes that we used
    delete @{$atts}{qw(required error valid regex notsession desc)};
}

#########################################################
sub textarea
{
    my($atts, $ret, $default)=@_;
    $default=~s/^\s+//;
    $default=~s/\s+$//;

    my $name=_clean_name($atts->{name});
    my $value=quotemeta($default||'');
    _perl_args(qq(\$$name=>"$value"));

    _std_var('$', $name, $atts);

    my $more=_nice_atts($atts);
    return qq(<TEXTAREA$more><%\$$name%></TEXTAREA>);
}

#########################################################
sub form
{
    my($atts, $ret)=@_;
    $session=lc $atts->{session};
    $session=0 if $session eq 'off' or $session eq 'no';
    $session_load='';
    $session_load=_nice_code($atts->{load}) if $atts->{load};
    $session_save='';
    $session_load=_nice_code($atts->{save}) if $atts->{save};

    $err_bad=$atts->{bad}           if yesno_att($atts, 'bad');
    $err_missing=$atts->{missing}   if yesno_att($atts, 'missing');

    delete @{$atts}{qw(session load save bad missing)};
    my $more=_nice_atts($atts);
    return qq(<%INIT>my \$BYPASS=0;\n</%INIT>\n<FORM$more>);
}

#########################################################
sub form_END
{
    my($atts, $ret)=@_;
    return $ret;
}

#########################################################
sub bypass
{
    my($atts, $ret)=@_;
    return qq(<%INIT>\$BYPASS=1 if($atts->{if});\n</%INIT>)         
                if $atts->{if};
    return qq(<%INIT>\$BYPASS=1 unless($atts->{unless});\n</%INIT>) 
                if $atts->{unless};
    die "<%BYPASS> requires if or unless attributes";
}



#########################################################
sub __next
{
    my($atts, $ret)=@_;

    my $href=$atts->{href};
    my $id=$atts->{session};
    unless($href) {
        DEBUG and warn "<%NEXT> requires an href attribute\n";
        die "<%NEXT> requires an href attribute\n"  unless $href;
        return '';
    }
    $id||='';
    $id=qq(\$u->query_form(\$u->query_form(), eval {return($id)}); die "Error in user code: \$@" if \$@;)
        if $id;
    
    my $sess='';
    $sess='&$SAVE_SESSION;' if $session;

    _tailing(<<MASON);
if((not \$FIRST and not \$VALIDATION_ERROR) or \$BYPASS)
{
    $sess
    my \$u=URI->new_abs("$href", 'http:'.\$r->parsed_uri->path());
    \$u->scheme('');
    $id
#    \$r->status(Apache::Contants::HTTP_REDIRECT());
    \$r->cgi_header_out(Location=>\$u->as_string());
    \$r->send_http_header();
    HTML::Mason::Commands::mc_suppress_http_header(1);
    return;
}
MASON
    return '';
}






#############################################################################
# Helper Functions

#########################################################
sub _clean_name
{
    my($t)=@_;
    $t=~s/\W/_/g;           # turn into a valid identifier
    $t="_$t" unless $t=~/^[_A-Za-z]/;
    return $t;
}

#########################################################
sub _nice_code
{
    my($code)=@_;
    $code.=';' unless $code=~/;\s*$/;
    return qq(eval {$code}; die "Error in user code: \$@\\n" if \$@;);
}

#########################################################
sub _nice_atts
{
    my($atts)=@_;
    return '' unless keys %$atts;
    return join ' ', '', map { defined($atts->{$_}) ? 
                                    qq($_="$atts->{$_}") : $_
                             } keys %$atts;
}

#########################################################
sub yesno_att
{
    my($atts, $key)=@_;
    return unless exists $atts->{$key};         # attribute not present
    return 1 unless defined($atts->{$key});     # <TAG att> equiv =1
    return 1 if $atts->{$key};
    return;
}


#########################################################
### STOLEN FROM HTML::SimpleParse
sub parse_atts             
{
    my($atts)=@_;
    my @returns;
        
    pos($atts) = 0;  # Make sure we start searching at the beginning of the string

    while (1) 
    {
        next if $atts =~ m/\G\s+/gc;  # Get rid of leading whitespace

        if ( $atts =~ m/\G
                (\S+?)=                                  # the key
                (?:
                 "([^\"\\]*  (?: \\.[^\"\\]* )* )"\s*    # quoted string, with possible whitespace inside,
                  |                                      #  or
                 ([^\s>]*)\s*                            # anything else, without whitespace or >
                )/gcx ) 
        {
            my ($key, $val) = ($1, $+);
            $val =~ s/\\(.)/$1/gs;
            push @returns, lc($key), $val;

        } elsif ( $atts =~ m/\G(\S+)\s*/gc ) 
        {
            push @returns, lc($1), undef;
        } else 
        {
            last;
        }
    }

    return @returns;
}




1;
__END__

=head1 NAME

HTML::Mason::Forms - Mason extension for luser-safe component syntax

=head1 SYNOPSIS

    use HTML::Mason::Forms;
    use HTML::Mason;
    my $parser = new HTML::Mason::Parser 
            (
                taint_check=>1, 
                'preprocess'=>\&HTML::Mason::Forms::preprocess,
                ....
            );


=head1 DESCRIPTION

HTML::Mason::Forms is a Mason pre-processor that adds some luser-safe 
HTML-like syntax to Mason components.

=over 4

=item E<lt>%MASON src="header.mason" title="Hello world" root=".."E<gt>

This is a wraper around mc_comp().  The SRC attribute is the component file
name, all other attributes are given as arguements to the component.  The
code above is equivalent to the following.

    %mc_comp("header.mason", title=>"Hello world", root=>"..");

=item E<lt>%INPUT type=input name="e-mail" value="your@email.com">

Short cut for text input.  This will output an E<lt>INPUT> that will take it's
current value from $ARGS, with the VALUE attribute as default. 
Additionnaly, the variable $e_mail is created via the E<lt>%args>

The code above is equivalent to the following:

    <%args>
    ${e-mail}="your@email.com"
    </%args>
    <INPUT type=input name="e-mail" value="<%$e_mail%>">

Note that the name "e-mail" is changed into a valid Perl identifier ($e_mail). 
And that $e_mail isn't HTML escaped.

=item E<lt>%INPUT type=passwd name=E<34>don't showE<34> value=E<34>your passwordE<34>>

Similar short cut as E<lt>INPUT type=text> above

The code above is equivalent to the following:

    <%args>
    ${don't show}="your password"
    </%args>
    <INPUT type=input name="don't show" value="<%$don_t_show%>">

=item E<lt>%INPUT type=checkbox name=meal value=supper>

Short cut for check boxes.  This will output an E<lt>INPUT> that will 
be CHECKED if it was set in $ARGS.  If you want it CHECKED as a default,
included the CHECKED attribute in the tag.
Additionnaly, the variable %meal is created via the E<lt>%args>

An example

    <%INPUT type=checkbox name=meal value="breakfast"> Breakfast<BR>
    <%INPUT type=checkbox name=meal value="lunch" CHECKED> Lunch<BR>
    <%INPUT type=checkbox name=meal value="tea"> Tea <BR>
    <%INPUT type=checkbox name=meal value="supper" CHECKED> Supper<BR>

Is equivalent to

    <%args>
    %meal=>(supper=>1, lunch=>1)
    </%args>
    <INPUT type=checkbox name="meal" value="breakfast" <%exists $meal{breakfast} ? ' CHECKED' : ''%>> Breakfast<BR>
    <INPUT type=checkbox name="meal" value="lunch" <%exists $meal{lunch} ? ' CHECKED' : ''%>> Lunch<BR>
    <INPUT type=checkbox name="meal" value="tea" <%exists $meal{tea} ? ' CHECKED' : ''%>> Tea<BR>
    <INPUT type=checkbox name="meal" value="supper" <%exists $meal{supper} ? ' CHECKED' : ''%>> Supper<BR>

=item E<lt>%INPUT type=radio name=city value="Montr&eacute;al">

Short cut for radio buttons.  This will output an E<lt>INPUT> that will 
be CHECKED if it was set in $ARGS.  If you want it CHECKED as a default,
included the CHECKED attribute in the tag.
If more then one CHECKED tag is present the first one is used.
Additionnaly, the variable $city is created via the E<lt>%args>

An example
    
    <%INPUT type=radio name=city value="mtl" CHECKED> 
    Montr&eacute;<BR>
    <%INPUT type=radio name=city value="qc"> 
    Qu&eacute;bec <BR>
    <%INPUT type=radio name=city value="lvl"> 
    Laval<BR>

Is equivalent to

    <%args>
    $city=>"mtl"
    </%args>
    <INPUT type=radio name=city value="mtl"<%($city||'') eq 'mtl' ? ' CHECKED ' : ''$>> 
    Montr&eacute;<BR>
    <INPUT type=radio name=city value="qc"<%($city||'') eq 'qc' ? ' CHECKED ' : ''$>> 
    Qu&eacute;bec <BR>
    <INPUT type=radio name=city value="lvl"<%($city||'') eq 'lvl' ? ' CHECKED ' : ''$>> 
    Laval<BR>

=head1 AUTHOR

Philip Gwyn, mason@artware.qc.ca

=head1 SEE ALSO

HTML::Mason, HTML::Mason::Parser

=cut

